#include "compilerApi.h"

#include "anls_att.h"

Processor_def::Processor_def()
{
  symb=NULL;
  dim=0;
}

void Processor_def::setsymbol(SgSymbol * sym)
{
  symb=sym;
}

SgSymbol * Processor_def::getsymbol()
{
  return (symb);
}

void Processor_def::setgrp(SgSymbol * sym)
{
  grp=sym;
}

SgSymbol * Processor_def::getgrp()
{
  return (grp);
}

void Processor_def::setshp(SgSymbol * sym)
{
  shp=sym;
}

SgSymbol * Processor_def::getshp()
{
  return (shp);
}

void Processor_def::add_dimention(int start, int end)
{
  dim++;
  if (dim>MAX_PROCESSOR_DIM)
    printf("Too many dims\n");
  s[dim-1]=start;
  e[dim-1]=end;
}

int Processor_def::getDim()
{
  return dim;
}

int Processor_def::getStart(int i)
{
  return s[i];
}

int Processor_def::getEnd(int i)
{
  return e[i];
}

void Processor_def::show()
{
  printf(" PR is %d dimention:\n   ", dim);
  if (symb!=NULL)
    printf("%s ", symb->identifier());
  for (int i=0; i< dim; i++) 
    printf("(%d, %d) ", s[i], e[i]);
  printf("\n");
}

//------------------------------;

Template_def::Template_def()
{
  symb=NULL;
  type=NULL;
  dim=0;
  lput=0;
}

void Template_def::setsymbol(SgSymbol * sym)
{
  symb=sym;
}

SgSymbol * Template_def::getsymbol()
{
  return (symb);
}

void Template_def::setrng(SgSymbol * sym)
{
  rng=sym;
}

SgSymbol * Template_def::getrng()
{
  return (rng);
}

void Template_def::settype(SgType * t)
{
  type=t;
}

SgType * Template_def::gettype()
{
  return (type);
}

void Template_def::add_dimention(int start, int end)
{
  dim++;
  if (dim>MAX_TEMPLATE_DIM)
    printf("Too many dims\n");
  s[dim-1]=start;
  e[dim-1]=end;
}

int Template_def::getDim()
{
  return dim;
}

int Template_def::getStart(int i)
{
  return s[i];
}

int Template_def::getEnd(int i)
{
  return e[i];
}

void Template_def::setlocalsize(int i, int size)
{
  localsize[i]=size;
}

int Template_def::getlocalsize(int i)
{
  return(localsize[i]);
}

void Template_def::setlput(int i)
{
  if (i>lput)
    lput=i;
}

int Template_def::getlput()
{
  return lput;
}

void Template_def::show()
{
  printf(" TP is %d dimention:\n   ", dim);
  if (symb!=NULL)
    printf("%s ", symb->identifier());

  int i;

  for (i=0; i< dim; i++) 
    printf("(%d, %d) ", s[i], e[i]);
  printf("\n");

  for (i=0; i< dim; i++) 
    printf("  [%d] ", localsize[i]);
  printf("  {%d}\n", lput);

}
//----------------------------;

Alignment_def::Alignment_def()
{
  dim=0;
  align_with_dim = 0;
}

void Alignment_def::setsymbol(SgSymbol * sym1, SgSymbol * sym2)
{
  fromsymb=sym1;
  tosymb=sym2;
  
  SgType * type1;
  type1=sym1->type()->internalBaseType();

  SgAttribute * a;
  a=sym2->getAttribute(0, Temp_Def);
  
  // if a is not a Template ??   zheng.gb
  if ( a==NULL) return;

  SgType * type2;
  type2 =  ((Template_def *)a->getAttributeData())->gettype();

  if (type2==NULL)
    ((Template_def *)a->getAttributeData())->settype(type1);
  else if (type1!=type2)
    printf("Only data with same type can be aligned to the same template\n");
}

SgSymbol * Alignment_def::getfromsymbol()
{
  return (fromsymb);
}

SgSymbol * Alignment_def::gettosymbol()
{
  return (tosymb);
}

void Alignment_def::setlocal(SgSymbol * sym)
{
  local=sym;
}

SgSymbol * Alignment_def::getlocal()
{
  return (local);
}

void Alignment_def::setdad(SgSymbol * sym)
{
  dad=sym;
}

SgSymbol * Alignment_def::getdad()
{
  return (dad);
}

void Alignment_def::settmb(SgSymbol * sym)
{
  tmb=sym;
}

SgSymbol * Alignment_def::gettmb()
{
  return (tmb);
}

void Alignment_def::settmbdad(SgSymbol * sym)
{
  tmbdad=sym;
}

SgSymbol * Alignment_def::gettmbdad()
{
  return (tmbdad);
}

void Alignment_def::add_dimention_info(SgExpression * s, SgExpression * o)
{
  dim++;
  if (dim>MAX_ALIGN_DIM)
    printf("Too many dims\n");
  stride[dim-1]=s;
  offset[dim-1]=o;
}

SgExpression * Alignment_def::getStride(int i)
{
  return stride[i];
}

int Alignment_def::getDim()
{
  return dim;
}

SgExpression * Alignment_def::getOffset(int i)
{
  return offset[i];
}

// done before add_dimention_info()
void Alignment_def::setAlignPair(int i)
{
	align_pair[dim]=i;
}

int Alignment_def::getAlignPair(int i)
{
	return align_pair[i];
}

void Alignment_def::setAlignWithDim(int i)
{
	align_with_dim=i;
}

int Alignment_def::getAlignWithDim()
{
	return align_with_dim;
}

void Alignment_def::show()
{
  printf(" AL is %d dimention:\n   ", dim);
  if ((fromsymb!=NULL)&&(tosymb!=NULL))
    printf("from %s to %s (%d dim)  ", fromsymb->identifier(), tosymb->identifier(), align_with_dim);
  for (int i=0; i< dim; i++) 
    printf("( %d [ %d, %d ]) ", align_pair[i], stride[i]->valueInteger(), offset[i]->valueInteger());
  printf("\n");
}

//--------------------------------------;

Distribute_def::Distribute_def()
{
  dim=0;
}

void Distribute_def::setsymbol(SgSymbol * sym1, SgSymbol * sym2)
{
  fromsymb=sym1;
  tosymb=sym2;
}

SgSymbol * Distribute_def::getfromsymbol()
{
  return (fromsymb);
}

SgSymbol * Distribute_def::gettosymbol()
{
  return (tosymb);
}

void Distribute_def::add_dimention_info(int s)
{
  dim++;
  if (dim>MAX_DISTRIBUTE_DIM)
    printf("Too many dims\n");
  distblock[dim-1]=s;
}

int Distribute_def::distCode(int i)
{
  return distblock[i];
}

void Distribute_def::show()
{
  printf(" DT is %d dimention:\n   ", dim);
  if ((fromsymb!=NULL)&&(tosymb!=NULL))
    printf("from %s to %s  ", fromsymb->identifier(), tosymb->identifier());
  for (int i=0; i< dim; i++) 
    printf("(%d) ", distblock[i]);
  printf("\n");
}

//----------------------------;;







