#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <string.h>

#include "compilerApi.h"
#include "anls_att.h"

extern void mark_no_align(SgStatement *);

extern void analyse_processor(SgStatement *);
extern void show_processor(SgStatement *);
extern void add_processor_declaration(SgStatement *);

extern void analyse_alignment(SgStatement *);
extern void show_alignment(SgStatement *);
extern void add_template_declaration(SgStatement *);

extern void analyse_template(SgStatement *);
extern void show_template(SgStatement *);

extern void analyse_distribute(SgStatement *);
extern void show_distribute(SgStatement *);

extern void analyse_inherit(SgStatement *);

extern void caculate_template_local_size(SgStatement *);

extern void add_align_array(SgStatement *);
extern void add_dummy_array(SgStatement *);

extern void analyse_forall_header(SgStatement *);
extern void get_forall_reference(SgStatement *);
extern void index_forall_reference(int, SgSymbol *, SgStatement *);
extern void detect_forall_reference(SgStatement *);
extern void show_forall_reference(SgStatement *);
extern void analyse_forall_template(SgStatement *);
extern void show_forall_template(SgStatement *);

extern void add_temp_base(SgStatement *);

extern void analyse_assign(SgStatement *);
extern void analyse_datatransfer(SgStatement *);

//------------------;

extern void add_definition(SgStatement *);
extern void add_new_dec(SgStatement *);

extern void decide_global_shadow(SgStatement *);
extern void delete_align_array(SgStatement *);
extern void delete_dummy_array(SgStatement *);

extern void trans_forall(SgStatement *);

extern void trans_datatransfer(SgStatement *);

extern void trans_assign(SgStatement *);

extern void trans_procstmt(SgStatement *);
extern void change_prochedr(SgStatement *);

char * mixstring="pcrc";

SgFile * global_file;

//flags like the one in unparse;

Flag in_main_program = off;
Flag in_procedure = off;

int do_level = 0;

void puttab(int number)
{
  for (int i=0; i<number; i++)
    printf("  ");
}

void tree_exp(SgExpression * expr, int level)
{
  static int tabnumber=0;

  level--;
  tabnumber++;
  if(level!=0) {
    puttab(tabnumber);
    if (expr!=NULL)
      printf("id: %3d, variant: %3d\n", expr->id(), expr->variant());
    else {
      printf("nil nil nil nil nil \n");
      tabnumber--;
      return;
    }

    if (expr->operand(1)!=NULL) {
      tree_exp(expr->operand(1), level);
    }
    else {
      tabnumber++;
      puttab(tabnumber);
      printf("nil nil nil nil nil \n");
      tabnumber--;
    };

    if (expr->operand(2)!=NULL)
      tree_exp(expr->operand(2), level);
    else {
      tabnumber++;
      puttab(tabnumber);
      printf("nil nil nil nil nil \n");
      tabnumber--;
    };
  }
  tabnumber--;
}

SgExpression * find_exp(int id)
{
  SgExpression * temp = global_file->firstExpression();
  printf("Finding exp %3d ...\n", id);
  for (; temp; temp=temp->nextInExprTable()) {
    if(temp->id()==id) 
      break;
   }
 return temp;
}

//------------------------;


int getreflevel(SgSymbol * index, SgExpression * expr)
{
  if (expr) {
    switch(expr->variant()) {

    case INT_VAL:
      return(0);
      break;

    case VAR_REF:
      if ((index)==(expr->symbol()))
	return (1);
      else
	return (0);
      break;

    case MINUS_OP:
    case ADD_OP:    
      {
	int left, right;
	left=getreflevel(index, expr->operand(1));
	right=getreflevel(index, expr->operand(2));
	
	return((left>right) ? left:right);
	break;
      };

    case MULT_OP:      
      {
	int left, right;
	left=getreflevel(index, expr->operand(1));
	right=getreflevel(index, expr->operand(2));
	
	return((left+right));
	break;
      };
                // zheng.gb
    case  FUNC_CALL:
    case  INTRIN_FUNC_CALL:
//      printf("getreflevel:function call.\n");
        return 0;
        break;
    default:
      printf("getreflevel::Expression not touched: %d\n", expr->id());
      return (100);
      break;
    };
  };

  return (-1);
}


// now we only have a simple to deal with s*index+o, and s!=0;
// so, i can only be 1 or 0;

SgExpression * getlinerexp(int i, SgSymbol * symb, SgExpression * expr)
{
  switch(expr->variant()) {
  case VAR_REF: {
      if (i==0) {
	return NULL;
      };
      if (i==1) {
	SgValueExp * stride = new SgValueExp(1);
	return stride;
      };
    break;
  };
   
  case MULT_OP: {
    if (i==0) {
      return NULL;
    };
    if (i==1) {
      int reflevel=getreflevel(symb, expr->operand(1));
      if (reflevel==1) 
	return expr->operand(2);
      else 
	return expr->operand(1);
    };
    break;
  }

  case ADD_OP: {
    if (i==0) {
      int reflevel=getreflevel(symb, expr->operand(1));
      if (reflevel==1) 
	return expr->operand(2);
      else 
	return expr->operand(1);
    };
    if (i==1) {
      int reflevel=getreflevel(symb, expr->operand(1));
      if (reflevel==1) 
	return (getlinerexp(1, symb, expr->operand(1)));
      else 
	return (getlinerexp(1, symb, expr->operand(2)));
    };
    break;
  }

  case  MINUS_OP: {
    if (i==0) {
      int reflevel=getreflevel(symb, expr->operand(1));
      if (reflevel==1) {
	SgExpression * minus = new SgExpression(MINUS_OP, NULL, NULL, NULL);
	minus->setLhs(expr->operand(2));
	return (minus);
      }
      else 
	return expr->operand(1);
    };
    if (i==1) {
      int reflevel=getreflevel(symb, expr->operand(1));
      if (reflevel==1) 
	return (getlinerexp(1, symb, expr->operand(1)));
      else {
	SgExpression * minus = new SgExpression(MINUS_OP, NULL, NULL, NULL);
	minus->setLhs(getlinerexp(1, symb, expr->operand(2)));
	return (minus);
      }
    };
    break;
  }
                // zheng.gb
  case  FUNC_CALL:
  case  INTRIN_FUNC_CALL:
        printf("getlinerexp:function call\n");
    if (i==0) {
      return NULL;
    };
        return 0;
  default:
    printf("Expression not touched!\n");
    break;
  };

  return NULL;
}

//------------------------;

SgExpression * forall_index(SgStatement * s)
{
  SgExpression * temp = s->expr(0);

  while (temp->variant()!=INDEX_OP)
    temp=temp->operand(1);
  //printf("We only deal one dimention forall\n");
  return (temp->operand(1));
}

void show_template_from_symbol(SgSymbol * symb)
{
    int i=0;

    for (symb; symb; symb=symb->next()) {
      if(symb->numberOfAttributes(Temp_Def)) {
	((Template_def *)(symb->getAttribute(0, Temp_Def)->getAttributeData()))->show();
      };
    };
}

SgType * gettype(int type)
{

  SgType * t = global_file->firstType();

  while (t!=NULL) {
    if (t->variant()==type)
      return t;
    t=t->next();
  }
  return (t);

}

//This functionis not well implemented, we should use more sophiscated SgVarDec...;

void add_var_dec(SgStatement * s, SgType * type, SgSymbol * name)
{

  SgExpression * expr = new SgExpression(VAR_REF, NULL, NULL, name);
  
  SgExpression * exprtype = new SgExpression(TYPE_OP, NULL, NULL, NULL, type);
  
  SgStatement * declstatement =new SgStatement(VAR_DECL, NULL, NULL, expr, exprtype, NULL);

  s->childList1(0)->insertStmtBefore(*declstatement, *s);

}

//--------------------------------;

char * int_typename="integer";
char * float_typename="real";

char * nameOfType(SgType * stype)
{    
  char * tname = NULL;

  if (stype->hasBaseType())
    stype=stype->baseType();

  switch (stype->variant()) {
  case T_BOOL:

    break;

  case T_CHAR:
    
    break;

  case T_INT:
    tname=int_typename;
    break;
  case T_FLOAT:
    tname=float_typename;
    break;
  case T_DOUBLE:
    
    break;
  case T_COMPLEX:
    
    break;
  case T_DCOMPLEX:
    
    break;
  case T_ARRAY:
    
    break;
  default:
    // do nothing;
    break;
  };

  return tname;
}
//--------------------------------;

char * int_allocname="integer_alloc";
char * float_allocname="real_alloc";

char * allocNameOfType(SgType * stype)
{    
  char * tname = NULL;

  if (stype->hasBaseType())
    stype=stype->baseType();

  switch (stype->variant()) {
  case T_BOOL:

    break;

  case T_CHAR:
    
    break;

  case T_INT:
    tname=int_allocname;
    break;
  case T_FLOAT:
    tname=float_allocname;
    break;
  case T_DOUBLE:
    
    break;
  case T_COMPLEX:
    
    break;
  case T_DCOMPLEX:
    
    break;
  case T_ARRAY:
    
    break;
  default:
    // do nothing;
    break;
  };

  return tname;
}
//--------------------------------;

char * int_freename="integer_free";
char * float_freename="real_free";

char * freeNameOfType(SgType * stype)
{    
  char * tname = NULL;

  if (stype->hasBaseType())
    stype=stype->baseType();

  switch (stype->variant()) {
  case T_BOOL:

    break;

  case T_CHAR:
    
    break;

  case T_INT:
    tname=int_freename;
    break;
  case T_FLOAT:
    tname=float_freename;
    break;
  case T_DOUBLE:
    
    break;
  case T_COMPLEX:
    
    break;
  case T_DCOMPLEX:
    
    break;
  case T_ARRAY:
    
    break;
  default:
    // do nothing;
    break;
  };

  return tname;
}

//--------------------------------;

char * int_typesizename="size_int";
char * float_typesizename="size_real";

char * sizeNameOfType(SgType * stype)
{    
  char * tname = NULL;

  if (stype->hasBaseType())
    stype=stype->baseType();

  switch (stype->variant()) {
  case T_BOOL:

    break;

  case T_CHAR:
    
    break;

  case T_INT:
    tname=int_typesizename;
    break;
  case T_FLOAT:
    tname=float_typesizename;
    break;
  case T_DOUBLE:
    
    break;
  case T_COMPLEX:
    
    break;
  case T_DCOMPLEX:
    
    break;
  case T_ARRAY:
    
    break;
  default:
    // do nothing;
    break;
  };

  return tname;
}
//--------------------------------;

char * int_typestackname="integer_stack";
char * float_typestackname="real_stack";

char * stackNameOfType(SgType * stype)
{    
  char * tname = NULL;

  if (stype->hasBaseType())
    stype=stype->baseType();

  switch (stype->variant()) {
  case T_BOOL:

    break;

  case T_CHAR:
    
    break;

  case T_INT:
    tname=int_typestackname;
    break;
  case T_FLOAT:
    tname=float_typestackname;
    break;
  case T_DOUBLE:
    
    break;
  case T_COMPLEX:
    
    break;
  case T_DCOMPLEX:
    
    break;
  case T_ARRAY:
    
    break;
  default:
    // do nothing;
    break;
  };

  return tname;
}

//-----------------------------------------;

void analyse(SgStatement * s)
{
  switch(s->variant()) {

  case GLOBAL:
    {
      int org_numberOfChildren=s->numberOfChildrenList1();
      SgStatement * * children = new SgStatement * [org_numberOfChildren];

      int i;

      for (i=0; i<org_numberOfChildren; i++)
	children[i]=s->childList1(i);
      for (i=0; i<org_numberOfChildren; i++)
	analyse(children[i]);
      
      delete [] children;

      break;
    };

  case PROG_HEDR:
    {
      int org_numberOfChildren=s->numberOfChildrenList1();
      SgStatement * * children = new SgStatement * [org_numberOfChildren];

      int i; 

      for (i=0; i<org_numberOfChildren; i++)
	children[i]=s->childList1(i);

      // do nothing?;
      in_main_program=on;

      for (i=0; i<org_numberOfChildren; i++)
	analyse(children[i]);
      
      delete [] children;

      in_main_program=off;

      break;

    };

  case PROC_HEDR:
    {
      SgProcHedrStmt * prochedr = (SgProcHedrStmt *)s;
      int org_numberOfChildren=s->numberOfChildrenList1();
      SgStatement * * children = new SgStatement * [org_numberOfChildren];

      int i;

      for (i=0; i<org_numberOfChildren; i++)
	children[i]=s->childList1(i);

      in_procedure=on;

      for (i=0; i<prochedr->numberOfParameters(); i++) {
	prochedr->parameter(i)->addAttribute(DUMP_Def, NULL, 0);
	prochedr->addAttribute(DUMP_Def, NULL, 0);
      }

      for (i=0; i<org_numberOfChildren; i++)
	analyse(children[i]);
      
      delete [] children;
      
      in_procedure = off;

      break;
    };

  case PROCESSORS_STMT:
    analyse_processor(s);
    break;

  case TEMPLATE_STMT:
    analyse_template(s);
    break;

  case ALIGN_DECL:
    analyse_alignment(s);
    break;
    
  case DISTRIBUTE_DECL:
    analyse_distribute(s);
    break;

  case INHERIT_DECL:
    analyse_inherit(s);
    break;

  case VAR_DECL:
    break;

  case INTERFACE_STMT:
    //do nothing;
    break;

  case FOR_NODE:
    {
      do_level++;

      int org_numberOfChildren=s->numberOfChildrenList1();

      SgStatement * * children = new SgStatement * [org_numberOfChildren];

      int i;

      for (i=0; i<org_numberOfChildren; i++)
	children[i]=s->childList1(i);

      for (i=0; i<org_numberOfChildren; i++)
	analyse(children[i]);
      
      delete [] children;
      
      do_level--;

      break;
    }

  case LOGIF_NODE:
  case ELSEIF_NODE:
  case LOGIF_STMT:
  case SWITCH_NODE:
  case CASE_NODE:
    {
      int org_numberOfChildren1=s->numberOfChildrenList1();

      SgStatement * * children1 = new SgStatement * [org_numberOfChildren1];

      int org_numberOfChildren2=s->numberOfChildrenList2();

      SgStatement * * children2 = new SgStatement * [org_numberOfChildren2];

      int i;

      for (i=0; i<org_numberOfChildren1; i++)
	children1[i]=s->childList1(i);

      for (i=0; i<org_numberOfChildren1; i++)
	analyse(children1[i]);
      
      delete [] children1;

      for (i=0; i<org_numberOfChildren2; i++)
	children2[i]=s->childList2(i);

      for (i=0; i<org_numberOfChildren2; i++)
	analyse(children2[i]);
      
      delete [] children2;

      break;

    }

  case ASSGOTO_NODE:
  case STOP_STAT:
    // do nothing;
    break;


  case PROC_STAT:
    //do nothing at presenttime. we will analyse INTERFACE later;
    break;

  case ASSIGN_STAT:
    analyse_assign(s);
    break;

  case FORALL_STMT:
    {       
      analyse_forall_header(s);
      get_forall_reference(s);

      Forall_header_def * fhdef=(Forall_header_def *) s->getAttribute(0, FHDER_Def)->getAttributeData();
    
      for (int i=0; i<fhdef->getindexnumber(); i++) {
	SgSymbol * index=fhdef->getindexsymb(i);
	index_forall_reference(i, index, s->childList1(0));
      };

      detect_forall_reference(s);
      break;
    }

  case PRINT_STAT:
  case WRITE_STAT:
    analyse_datatransfer(s);
    break;

  case CONTROL_END:
  case PARAM_DECL:
  case ENDDO_STMT:
  case  FORMAT_STAT:
   // statement need do nothing;
    break;

  default:
    printf("Not touched statement in analyse on line %d\n", s->lineNumber());
    break;
  };

}

void transform_1(SgStatement * s) 
{
  switch(s->variant()) {
  case GLOBAL:
    {
      int org_numberOfChildren=s->numberOfChildrenList1();
      SgStatement * * children = new SgStatement * [org_numberOfChildren];
      
      int i;

      for (i=0; i<org_numberOfChildren; i++)
	children[i]=s->childList1(i);
      for (i=0; i<org_numberOfChildren; i++)
	transform_1(children[i]);
      
      delete [] children;
      break;
    };

  case PROG_HEDR:
    {
      in_main_program = on;

      // first remember the thing to do;
      int org_numberOfChildren=s->numberOfChildrenList1();
      SgStatement * * children = new SgStatement * [org_numberOfChildren];

      int i;

      for (i=0; i<org_numberOfChildren; i++)
	children[i]=s->childList1(i);

      // second, marked all those non explicitly distributed dada before new symbols are added;
      //mark_no_align(s);

      add_processor_declaration(s);
      add_template_declaration(s);

      caculate_template_local_size(s);

      decide_global_shadow(s);
      
      add_align_array(s);                  //   VAR_DECL of array

      //show_processor(s);
      //show_template(s);
      
      add_definition(s);
      
      for (i=0; i<org_numberOfChildren; i++)
	transform_1(children[i]);
      
      delete [] children;
      
      in_main_program=off;

      break;
    }

  case PROC_HEDR:
    {
      SgProcHedrStmt * prochedr = (SgProcHedrStmt *)s;
      int org_numberOfChildren=s->numberOfChildrenList1();
      SgStatement * * children = new SgStatement * [org_numberOfChildren];

      int i;

      for (i=0; i<org_numberOfChildren; i++)
	children[i]=s->childList1(i);

      in_procedure=on;

      add_processor_declaration(s);
      add_template_declaration(s); 
     
      caculate_template_local_size(s);

      add_align_array(s);
      add_dummy_array(s);

      add_definition(s);
      
      change_prochedr(s);

      for (i=0; i<org_numberOfChildren; i++)
	transform_1(children[i]);
      delete [] children;
      
      in_procedure = off;
      break;
    };

  case INTERFACE_STMT:
    s->deleteStmt();
    break;

  case PROCESSORS_STMT:
    //printf("- Processor Statement's id: % 4d variant:  % 4d\n", s->id(), s->variant());
    s->deleteStmt();
    break;

  case TEMPLATE_STMT:
    //printf("- Template Statement's  id: % 4d variant:  % 4d\n", s->id(), s->variant());
    s->deleteStmt();
    break;

  case ALIGN_DECL:
    //printf("- Align Statement's     id: % 4d variant:  % 4d\n", s->id(), s->variant());
    s->deleteStmt();
    break;

  case DISTRIBUTE_DECL:
    s->deleteStmt();
    break;

  case INHERIT_DECL:
    s->deleteStmt();
    break;

  case VAR_DECL:
    delete_align_array(s); 

    if (in_procedure==on)
      delete_dummy_array(s);

    break;

  case FOR_NODE:
    {
      do_level++;

      int org_numberOfChildren=s->numberOfChildrenList1();

      SgStatement * * children = new SgStatement * [org_numberOfChildren];

      int i;

      for (i=0; i<org_numberOfChildren; i++)
	children[i]=s->childList1(i);

      for (i=0; i<org_numberOfChildren; i++)
	transform_1(children[i]);
      
      delete [] children;
      
      do_level--;

      break;
    }

  case LOGIF_NODE:
  case ELSEIF_NODE:
  case LOGIF_STMT:
  case SWITCH_NODE:
  case CASE_NODE:
    {
      int org_numberOfChildren1=s->numberOfChildrenList1();

      SgStatement * * children1 = new SgStatement * [org_numberOfChildren1];

      int org_numberOfChildren2=s->numberOfChildrenList2();

      SgStatement * * children2 = new SgStatement * [org_numberOfChildren2];

      int i;

      for (i=0; i<org_numberOfChildren1; i++)
	children1[i]=s->childList1(i);

      for (i=0; i<org_numberOfChildren1; i++)
	transform_1(children1[i]);
      
      delete [] children1;

      for (i=0; i<org_numberOfChildren2; i++)
	children2[i]=s->childList2(i);

      for (i=0; i<org_numberOfChildren2; i++)
	transform_1(children2[i]);
      
      delete [] children2;

      break;

    }

  case ASSGOTO_NODE:
  case STOP_STAT:
    // do nothing;
    break;

  case PROC_STAT:
    trans_procstmt(s);
    s->deleteStmt();
    break;

  case ASSIGN_STAT:
    trans_assign(s);
    break;

  case FORALL_STMT:
    //show_forall_reference(s);
    trans_forall(s);
    s->deleteStmt();
    break;

  case PRINT_STAT:
  case WRITE_STAT:
    trans_datatransfer(s);
    s->deleteStmt();
    break;

  case CONTROL_END:
  case PARAM_DECL:
  case ENDDO_STMT:
  case  FORMAT_STAT:
    // statement need do nothing;
    break;

  default:
    printf("In transform_1, unexpected statement on line %d, id: %d, variant: %d\n", s->lineNumber(), s->id(), s->variant());
    break;
  };
}

void transform_2(SgStatement * s) 
{
  switch(s->variant()) {
  case GLOBAL:
    {
      int org_numberOfChildren=s->numberOfChildrenList1();
      SgStatement * * children = new SgStatement * [org_numberOfChildren];

      int i;

      for (i=0; i<org_numberOfChildren; i++)
	children[i]=s->childList1(i);
      for (i=0; i<org_numberOfChildren; i++)
	transform_2(children[i]);
      
      delete [] children;

      break;
    };

  case PROG_HEDR:
    {
      in_main_program = on;
      add_new_dec(s);
      in_main_program = off;

      break;
    };

  case PROC_HEDR:
    {
      in_procedure = on;
      add_new_dec(s);      
      in_procedure = off;

      break;
    };

  default:
    printf("Not touched statement in transform_2 on line %d\n", s->lineNumber());
    break;
  };
}

void test_transform(SgStatement * s) 
{
  while (s!=NULL) {
    if (s->variant()==ASSIGN_STAT) {
	printf("The right is int %d, %d\n",s->expr(1)->isInteger(), s->expr(1)->valueInteger());
    }
    
    s=s->lexNext();
  }
  
  s=global_file->firstStatement();

  SgLabel * label=global_file->firstLabel();

  while (label!=NULL) {
    printf("the lable id is: %d, %d\n", label->id(), label->labelValue());
    label=label->nextInScope(s);
  }
}

char *newOutputFileName(char *inputFileName)
{	
  char	*out;
  int	nameLength = strlen(inputFileName);
  
  out=(char *)malloc((unsigned)(nameLength+5));

  while (inputFileName[nameLength-1]!='.') {
    nameLength --;
  }

  for ( int i = 0; i<nameLength; i++) {
    out[i]=inputFileName[i];
  }

  out[nameLength-1]='\0';

  out=strcat(out,"_0.dep");

  return(out);
}

main(int argc, char * argv[]) 
{
  SgProject * project = new SgProject("test", argc, argv);

  for (int i=0; i< project->numberOfFiles(); i++) {

    global_file = &(project->file(i));

    SgStatement * s;

    s = global_file->firstStatement();

    //test_transform(s);
    
    analyse(s);

    printf("\n-----------------\n"); 

    transform_1(s);
    transform_2(s);

    global_file->saveDepFile(newOutputFileName(global_file->filept->filename));

  };

  return 0;
}






