//			dist2whole.C
//
//		complete_distribute_elements()
//
// written by Zheng gengbin

/*
This pass complete the distribute direvtives. Add default format or processor
if they missed .
*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#else
#include <strings.h>
#endif

#include "compilerApi.h"
#include "pass.h"
#include "ext_func.h"

extern int num_procs;

// get the processor whose dimension is dim0 , and size is size0
SgSymbol *getProc(SgStatement *hedr_stmt , int dim0 ,int size0)
{
//printf("dim0:%d , size0:%d\n",dim0,size0);
	for (int i=0 ; i< hedr_stmt->numberOfAttributes(Proc_Def) ; i++)
	{
		SgSymbol     *proc  = (SgSymbol *)hedr_stmt->attributeValue(i,Proc_Def);
		int dim  = ((SgArrayType*)proc->type())->dimension();
		int size = ((SgArrayType*)proc->type())->getSize();
//printf("proc:%s,dim:%d , size:%d\n",proc->identifier(),dim,size);
		if ( dim == dim0  && size == size0 ) return proc;
	}
	return NULL;
}

// pow() always cause problem when link(-lm) , so write one myself.
inline int pow0(int x,int y)
{
        if (y==0) return 1;
        return x*pow0(x,y-1);
}

int devide_dim_proc(int nproc,int dim , int (&dimarray)[8])
{
	int i;

	if (dim == 0)   return 0;
	if (nproc < 2) return 0;
	if (dim == 1)  { dimarray[0] = nproc ; return 1; }

	// find a better way to divide balantly : pow(i,dim) = nproc
	for (i=1; i<=nproc/2 ; i++)
		if (pow0(i,dim) == nproc)
		{
			for (int j = 0; j<dim ; j++ ) dimarray[j] = i;
			return 1;
		}
	// so have to try one by one
	for (i = 2; i<nproc ; i++)
	{
		if (nproc % i != 0) continue;
		dimarray[dim-1]=i; 
		if (!devide_dim_proc(nproc/i,dim-1,dimarray)) continue;
		else break;
	}
	if (i==nproc) return 0;
	else return 1;
}

SgSymbol  *get_dim_processor(SgStatement *hedr_stmt,int dim)
{
	int variant = hedr_stmt->variant();
	ASSERT(variant == PROG_HEDR || variant == PROC_HEDR || variant == FUNC_HEDR );

	// if present a default one 
//	if (defaultProcs[dim]) return defaultProcs[dim];
	SgSymbol *proc = getProc(hedr_stmt, dim , num_procs);
	if (proc) return proc;

	proc = get_temp_variable(hedr_stmt,"PROC");
	SgType   *type = SgTypeProcessors();
	int       dimarray[8];

	if (!devide_dim_proc(num_procs,dim,dimarray))
		return NULL;

	SgExpression *expr_list=NULL;
	for (int i =0; i< dim ; i++)
	{
		SgExpression *ddot=new SgSubscriptExp(*new SgValueExp(1),*new SgValueExp(dimarray[i]));
		expr_list = new SgExpression(EXPR_LIST,ddot,expr_list,NULL);
	}
	SgArrayType *arraytype=new SgArrayType(*type,dim,*expr_list);
	proc->setType(arraytype);

	SgExpression *arrayref=new SgArrayRefExp(*proc,*expr_list);
//	SgExpression *ref = new SgExpression(EXPR_LIST,arrayref,NULL,NULL); 
	SgExpression *ref = arrayref;

	SgStatement *decl=new SgVarDeclStmt(* ref,* type);
        decl->setVariant(PROCESSORS_STMT);
        hedr_stmt->firstStmtForDeclaration()->insertStmtAfter(*decl);
	// add attribute of Proc_Def
	hedr_stmt->addAttribute(Proc_Def , proc , sizeof(proc) );
	
//	defaultProcs[dim]=proc;
	return proc;
}

void complete_distribute_elements(SgStatement *s)
{
	SgExpression *dist_op , *dist_format , *dist_target;
	SgSymbol     *distributee;

	dist_op = s->expr(2)->lhs();
	dist_format = dist_op->lhs();
	dist_target = dist_op->rhs();
	distributee=s->expr(0)->lhs()->symbol();

	// if the distribute format is missed
	if (dist_format == NULL)
	{
	    if (distributee->type()->variant() == T_ARRAY) 
	    {
		SgArrayType  *atype = (SgArrayType *)(distributee->type());
		for (int i=0; i<atype->dimension() ; i++ )
		{
			SgExpression *node=new SgExpression(BLOCK_OP,NULL,NULL,NULL);
			dist_format = new SgExpression(EXPR_LIST,node,dist_format,NULL);
		}
		dist_format = new SgExpression(DIST_FORMAT_OP,NULL,dist_format,NULL);
		dist_op->setLhs(dist_format);
	    }
	}

	// or if dist_target is missed
	if (dist_target == NULL)
	{
	    SgStatement *hedr_stmt = s->getScopeForDeclare();
		// got the desire dimension, don't consider the * in dist_format
	    int dim=0;
	    for (SgExpression *expr_list=dist_format->rhs() ; expr_list ;
			expr_list=expr_list->rhs())
		if (expr_list->lhs()->variant() != STAR_OP) dim++;

		// get the processors if possible !
	    SgSymbol *procs=get_dim_processor(hedr_stmt,dim);
	    if (procs == NULL)
	    {
		printf("Can not handled to devide the %d procesors in %d dims ,aborted !\n",num_procs,dim);
                exit(1);
	    }

	    SgExpression *ref=new SgVarRefExp(procs);
	    dist_target = new SgExpression(DIST_TARGET_OP , NULL, ref,NULL);
	    dist_op->setRhs(dist_target);
	}
}


