
#include 	<stdio.h>
#include        <stdlib.h>
#include 	<malloc.h>
#include	<string.h>

#include        "compilerApi.h"

#include	"basic_fun.h"
#include	"global_var.h"
#include        "trans2.h"

//char	*get_file_out(char *);
char *newOutputFileName(char *inputFileName);
int array_to_forall(SgStatement *,SgStatement *);
int element_to_array(SgStatement *,SgStatement *);
int where_to_forall(SgStatement *,SgStatement *);

SgVariableSymb	*temp_index[8];
SgValueExp 	*base_step;//default step is 1
SgValueExp	*base_lbound;
SgValueExp	*default_lsub;//Fortran default low bound is 1
SgStatement     *current;//current statement to be transformed
int		index_g;//digit for forall index variable. 
int		index_var;//digit for temporary variable.

PassType passtype;

enum Flag { off, on };
int in_main_program = off;
int in_procedure = off;

int do_level = 0;

/* This function is used to cut the surfix, so there is a record as different
pass performed */
char *newOutputFileName(char *inputFileName)
{
  char  *out;
  int   nameLength = strlen(inputFileName);
  
  out=(char *)malloc((unsigned)(nameLength+5));

  while (inputFileName[nameLength-1]!='.') {
    nameLength --;
  }

  for ( int i = 0; i<nameLength; i++) {
    out[i]=inputFileName[i];
  }

  out[nameLength-1]='\0';

  out=strcat(out,"_2.dep");

  return(out);
}

void analyse(SgStatement * s)
{
  switch(s->variant()) {

  case GLOBAL:
    {
      int org_numberOfChildren=s->numberOfChildrenList1();
      SgStatement * * children = new SgStatement * [org_numberOfChildren];

      int i;

      for (i=0; i<org_numberOfChildren; i++)
	children[i]=s->childList1(i);
      for (i=0; i<org_numberOfChildren; i++)
	analyse(children[i]);
      
      delete [] children;

      break;
    };

  case PROG_HEDR:
    {
      int org_numberOfChildren=s->numberOfChildrenList1();
      SgStatement * * children = new SgStatement * [org_numberOfChildren];

      int i; 

      for (i=0; i<org_numberOfChildren; i++)
	children[i]=s->childList1(i);

      // do nothing?;
      in_main_program=on;

	// do some initialization 
      init_temp_index(s);

      for (i=0; i<org_numberOfChildren; i++)
	analyse(children[i]);
      
      delete [] children;

      in_main_program=off;
	
      break;

    };

  case PROC_HEDR:
    {
      SgProcHedrStmt * prochedr = (SgProcHedrStmt *)s;
      int org_numberOfChildren=s->numberOfChildrenList1();
      SgStatement * * children = new SgStatement * [org_numberOfChildren];

      int i;

      for (i=0; i<org_numberOfChildren; i++)
	children[i]=s->childList1(i);

      in_procedure=on;
	// do some initialization 
      init_temp_index(s);

      for (i=0; i<org_numberOfChildren; i++)
	analyse(children[i]);
      
      delete [] children;
      
      in_procedure = off;

      break;
    };

  case FUNC_HEDR:
    {
      SgProcHedrStmt * prochedr = (SgProcHedrStmt *)s;
      int org_numberOfChildren=s->numberOfChildrenList1();
      SgStatement * * children = new SgStatement * [org_numberOfChildren];

      int i;

      for (i=0; i<org_numberOfChildren; i++)
	children[i]=s->childList1(i);

      in_procedure=on;
	// do some initialization 
      init_temp_index(s);

      for (i=0; i<org_numberOfChildren; i++)
	analyse(children[i]);
      
      delete [] children;
      
      in_procedure = off;

      break;
    };

  case FOR_NODE:
    {
      do_level++;

      int org_numberOfChildren=s->numberOfChildrenList1();

      SgStatement * * children = new SgStatement * [org_numberOfChildren];

      int i;

      for (i=0; i<org_numberOfChildren; i++)
	children[i]=s->childList1(i);

      for (i=0; i<org_numberOfChildren; i++)
	analyse(children[i]);
      
      delete [] children;
      
      do_level--;

      break;
    }

  case LOGIF_NODE:
  case ELSEIF_NODE:
  case LOGIF_STMT:
  case SWITCH_NODE:
  case CASE_NODE:
    {
      int org_numberOfChildren1=s->numberOfChildrenList1();

      SgStatement * * children1 = new SgStatement * [org_numberOfChildren1];

      int org_numberOfChildren2=s->numberOfChildrenList2();

      SgStatement * * children2 = new SgStatement * [org_numberOfChildren2];

      int i;

      for (i=0; i<org_numberOfChildren1; i++)
	children1[i]=s->childList1(i);

      for (i=0; i<org_numberOfChildren1; i++)
	analyse(children1[i]);
      
      delete [] children1;

      for (i=0; i<org_numberOfChildren2; i++)
	children2[i]=s->childList2(i);			// zheng.gb

      for (i=0; i<org_numberOfChildren2; i++)
	analyse(children2[i]);
      
      delete [] children2;

      break;

    }

  case ASSIGN_STAT:
    {
	// change array element to array section like: a(1) => a(1:1)
	// this will leave the assign of array element to trans0
	if (passtype == ASSIGN_PASS)
	{
	SgStatement *hedr_stmt = s->getScopeForDeclare();
	ASSERT(hedr_stmt);
	element_to_array(hedr_stmt,s);
        array_to_forall(hedr_stmt,s);
	}
        break;
    }

  case WHERE_NODE:
    {
	if (passtype == WHERE_STMT_PASS)
	{
	SgStatement *hedr_stmt = s->getScopeForDeclare();
	ASSERT(hedr_stmt);
        where_to_forall(hedr_stmt,s);
	}
        break;
    }

  case WHERE_BLOCK_STMT:
    {
extern int whereconstruct_to_where(SgStatement *,SgStatement *);
	if (passtype == WHERE_CONSTRUCT_PASS)
	{
	SgStatement *hedr_stmt = s->getScopeForDeclare();
	ASSERT(hedr_stmt);
        whereconstruct_to_where(hedr_stmt,s);
	}
        break;
    }

  case FORALL_STMT:
    {       
      break;
    }

  case PRINT_STAT:
  case WRITE_STAT:
//    analyse_datatransfer(s);
    break;

  case CONTROL_END:
  case PARAM_DECL:
  case ENDDO_STMT:
  case  FORMAT_STAT:
   // statement need do nothing;
    break;

  default:
//    printf("Not touched statement in analyse on line %d\n", s->lineNumber());
    break;
  };

}


int main(int argc, char **argv)
{

//  	SgProject 	project("test", argc, argv);
	SgProject       *project;
 	SgFile	*file=NULL;

  	SgStatement	*hedr_stmt,*next_stmt,*end_stmt;
	int	i=0,numoffunc,numoffiles;
	char	*file_in,*file_out;

	PassType   pt[10];
	int no=0;

	argv++;
        argc--;
        while ((argc > 0) && ((*argv)[0] == '-') )
        {
                i=1;
                while (argv[0][i] != '\0') {
                        switch (argv[0][i]) {
                        case 'a'  :
				pt[no++]=WHERE_CONSTRUCT_PASS;
				break;
                        case 'b'  :
				pt[no++]=WHERE_STMT_PASS;
				break;
                        case 'c'  :
				pt[no++]=ASSIGN_PASS;
				break;
			default:
        			printf("warning 422: Unknown option \"%c\" ignored.\n",argv[0][i]);
        			exit(1);
			}
			i++;
		}
		argc--;
                argv++;
	}
	if ( no == 0 )
	{
		pt[no++]=WHERE_CONSTRUCT_PASS;
		pt[no++]=WHERE_STMT_PASS;
		pt[no++]=ASSIGN_PASS;
	}

	argc++; argv--;
	project = new SgProject("test", argc, argv);

	numoffiles=project->numberOfFiles();
	for(i=0;i<numoffiles;i++)
	{
		base_lbound=new SgValueExp(0);
		base_step=new SgValueExp(1);
		default_lsub=new SgValueExp(1);//fortran default low bound is 1
		file=&(project->file(i));	
		file_in=file->filept->filename;
  		if (file->languageType()!=ForSrc)
		{	printf("%s is  not fortran program.\n",file_in); 
			continue;
		}

		char comment[100];
		for (int j=0; j<no ; j++)
    		{
			passtype=pt[j];
			switch (passtype) {
        		case WHERE_CONSTRUCT_PASS :
                		sprintf(comment,"%s","Where construct ");
                		break;
        		case WHERE_STMT_PASS :
                		sprintf(comment,"%s","Where statement ");
                		break;
        		case ASSIGN_PASS :
                		sprintf(comment,"%s","Assign statement");
                		break;
			}
			printf("***********  PASS %s **********\n",comment);
			SgStatement *s = file->firstStatement();
			analyse(s);
		}

		file_out=newOutputFileName(file_in);
		file->saveDepFile(file_out);
		//save the transformation result (AST) to file result*.dep

	}
  	return 0;
}

